import random

import numpy as np

from algorithm.pla.perceptron import Perceptron


def train(perceptron: Perceptron, training_data: list, max_iterations: int, learning_rate: float):
    """ Trainiert ein Perzeptron nach dem PLA-Algorithmus """

    # Wiederhole bis keine Fehler gemacht werden oder die maximale Anzahl an Iterationen erreicht wurden
    for i in range(max_iterations):

        # Durchlaufe alle Testdatenpaare. Ein Paar besteht aus einem Merkmalsvektor und einer korrekten Klassifizierung
        wrong_classifications = []
        for features, correct_class in training_data:
            result = perceptron.classify(features)

            # Falls das Ergebnis falsch ist, merken wir uns die Ergebnisse
            if result != correct_class:
                wrong_classifications.append((features, correct_class, result))

        # Keine Fehler? Wir können aufhören mit trainieren
        if len(wrong_classifications) == 0:
            break

        # Doch Fehler? Wähle eine zufällige falsche Klassifizierung und update die Gewichte
        features, correct_class, result = random.choice(wrong_classifications)
        update_weight(perceptron, features, correct_class, result, learning_rate)


def train_pocket(perceptron: Perceptron, training_data: list, max_iterations: int, learning_rate: float):
    """ Trainiert ein Perzeptron nach dem Pocket-Algorithmus """

    # Initialisiere den Zwischenspeicher ("Pocket") für das beste Ergebnis
    min_false_classifications = len(training_data)
    best_weights = []

    # Wiederhole bis keine Fehler gemacht werden oder die maximale Anzahl an Iterationen erreicht wurden
    for i in range(max_iterations):

        # Durchlaufe alle Testdatenpaare. Ein Paar besteht aus einem Merkmalsvektor und einer korrekten Klassifizierung
        wrong_classifications = []
        for features, correct_class in training_data:
            result = perceptron.classify(features)

            # Falls das Ergebnis falsch ist, merken wir uns die Ergebnisse
            if result != correct_class:
                wrong_classifications.append((features, correct_class, result))

        # Wir sind alle Testdaten durchlaufen und haben eine minimale Fehlerrate?
        # → Wir merken uns die Gewichte
        if len(wrong_classifications) > min_false_classifications:
            min_false_classifications = len(wrong_classifications)
            best_weights = perceptron.weights

        # Keine Fehler? Wir können aufhören mit trainieren
        if len(wrong_classifications) == 0:
            break

        # Doch Fehler? Wähle eine zufällige falsche Klassifizierung und update die Gewichte
        features, correct_class, result = random.choice(wrong_classifications)
        update_weight(perceptron, features, correct_class, result, learning_rate)

    perceptron.weights = best_weights


def update_weight(perceptron, features, correct_class, classification, learning_rate):
    correction_factor = (correct_class - classification)
    multiplied_features = learning_rate * correction_factor * np.array([1] + features)
    new_weights = np.add(multiplied_features, np.array(perceptron.weights))
    perceptron.weights = new_weights.tolist()
